# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import fixtures
import mock
import oslo_messaging as messaging

from ironic_inspector.common import swift
from ironic_inspector.conductor import manager
import ironic_inspector.conf
from ironic_inspector import introspect
from ironic_inspector import process
from ironic_inspector.test import base as test_base
from ironic_inspector import utils

CONF = ironic_inspector.conf.CONF


class BaseManagerTest(test_base.NodeTest):
    def setUp(self):
        super(BaseManagerTest, self).setUp()
        self.mock_log = self.useFixture(fixtures.MockPatchObject(
            manager, 'LOG')).mock
        self.mock__shutting_down = (self.useFixture(fixtures.MockPatchObject(
            manager.semaphore, 'Semaphore', autospec=True))
            .mock.return_value)
        self.mock__shutting_down.acquire.return_value = True
        self.manager = manager.ConductorManager()
        self.context = {}
        self.token = None


class TestManagerInitHost(BaseManagerTest):
    def setUp(self):
        super(TestManagerInitHost, self).setUp()
        self.mock_db_init = self.useFixture(fixtures.MockPatchObject(
            manager.db, 'init')).mock
        self.mock_validate_processing_hooks = self.useFixture(
            fixtures.MockPatchObject(manager.plugins_base,
                                     'validate_processing_hooks')).mock
        self.mock_filter = self.useFixture(fixtures.MockPatchObject(
            manager.pxe_filter, 'driver')).mock.return_value
        self.mock_periodic = self.useFixture(fixtures.MockPatchObject(
            manager.periodics, 'periodic')).mock
        self.mock_PeriodicWorker = self.useFixture(fixtures.MockPatchObject(
            manager.periodics, 'PeriodicWorker')).mock
        self.mock_executor = self.useFixture(fixtures.MockPatchObject(
            manager.utils, 'executor')).mock
        self.mock_ExistingExecutor = self.useFixture(fixtures.MockPatchObject(
            manager.periodics, 'ExistingExecutor')).mock
        self.mock_exit = self.useFixture(fixtures.MockPatchObject(
            manager.sys, 'exit')).mock

    def assert_periodics(self):
        outer_cleanup_decorator_call = mock.call(
            spacing=CONF.clean_up_period)
        self.mock_periodic.assert_has_calls([
            outer_cleanup_decorator_call,
            mock.call()(manager.periodic_clean_up)])

        inner_decorator = self.mock_periodic.return_value
        inner_cleanup_decorator_call = mock.call(
            manager.periodic_clean_up)
        inner_decorator.assert_has_calls([inner_cleanup_decorator_call])

        self.mock_ExistingExecutor.assert_called_once_with(
            self.mock_executor.return_value)

        periodic_worker = self.mock_PeriodicWorker.return_value

        periodic_sync = self.mock_filter.get_periodic_sync_task.return_value
        callables = [(periodic_sync, None, None),
                     (inner_decorator.return_value, None, None)]
        self.mock_PeriodicWorker.assert_called_once_with(
            callables=callables,
            executor_factory=self.mock_ExistingExecutor.return_value,
            on_failure=self.manager._periodics_watchdog)
        self.assertIs(periodic_worker, self.manager._periodics_worker)

        self.mock_executor.return_value.submit.assert_called_once_with(
            self.manager._periodics_worker.start)

    def test_no_introspection_data_store(self):
        CONF.set_override('store_data', 'none', 'processing')
        self.manager.init_host()
        self.mock_log.warning.assert_called_once_with(
            'Introspection data will not be stored. Change "[processing] '
            'store_data" option if this is not the desired behavior')

    def test_init_host(self):
        self.manager.init_host()
        self.mock_db_init.asset_called_once_with()
        self.mock_validate_processing_hooks.assert_called_once_with()
        self.mock_filter.init_filter.assert_called_once_with()
        self.assert_periodics()

    def test_init_host_validate_processing_hooks_exception(self):
        class MyError(Exception):
            pass

        error = MyError('Oops!')
        self.mock_validate_processing_hooks.side_effect = error

        # NOTE(milan): have to stop executing the test case at this point to
        # simulate a real sys.exit() call
        self.mock_exit.side_effect = SystemExit('Stop!')
        self.assertRaisesRegex(SystemExit, 'Stop!', self.manager.init_host)

        self.mock_db_init.assert_called_once_with()
        self.mock_log.critical.assert_called_once_with(str(error))
        self.mock_exit.assert_called_once_with(1)
        self.mock_filter.init_filter.assert_not_called()


class TestManagerDelHost(BaseManagerTest):
    def setUp(self):
        super(TestManagerDelHost, self).setUp()
        self.mock_filter = self.useFixture(fixtures.MockPatchObject(
            manager.pxe_filter, 'driver')).mock.return_value
        self.mock_executor = mock.Mock()
        self.mock_executor.alive = True
        self.mock_get_executor = self.useFixture(fixtures.MockPatchObject(
            manager.utils, 'executor')).mock
        self.mock_get_executor.return_value = self.mock_executor
        self.mock__periodic_worker = self.useFixture(fixtures.MockPatchObject(
            self.manager, '_periodics_worker')).mock
        self.mock_exit = self.useFixture(fixtures.MockPatchObject(
            manager.sys, 'exit')).mock

    def test_del_host(self):
        self.manager.del_host()

        self.mock__shutting_down.acquire.assert_called_once_with(
            blocking=False)
        self.mock__periodic_worker.stop.assert_called_once_with()
        self.mock__periodic_worker.wait.assert_called_once_with()
        self.assertIsNone(self.manager._periodics_worker)
        self.mock_executor.shutdown.assert_called_once_with(wait=True)
        self.mock_filter.tear_down_filter.assert_called_once_with()
        self.mock__shutting_down.release.assert_called_once_with()

    def test_del_host_race(self):
        self.mock__shutting_down.acquire.return_value = False

        self.manager.del_host()

        self.mock__shutting_down.acquire.assert_called_once_with(
            blocking=False)
        self.mock_log.warning.assert_called_once_with(
            'Attempted to shut down while already shutting down')
        self.mock__periodic_worker.stop.assert_not_called()
        self.mock__periodic_worker.wait.assert_not_called()
        self.assertIs(self.mock__periodic_worker,
                      self.manager._periodics_worker)
        self.mock_executor.shutdown.assert_not_called()
        self.mock_filter.tear_down_filter.assert_not_called()
        self.mock__shutting_down.release.assert_not_called()
        self.mock_exit.assert_not_called()

    def test_del_host_worker_exception(self):
        class MyError(Exception):
            pass

        error = MyError('Oops!')
        self.mock__periodic_worker.wait.side_effect = error

        self.manager.del_host()

        self.mock__shutting_down.acquire.assert_called_once_with(
            blocking=False)
        self.mock__periodic_worker.stop.assert_called_once_with()
        self.mock__periodic_worker.wait.assert_called_once_with()
        self.mock_log.exception.assert_called_once_with(
            'Service error occurred when stopping periodic workers. Error: %s',
            error)
        self.assertIsNone(self.manager._periodics_worker)
        self.mock_executor.shutdown.assert_called_once_with(wait=True)
        self.mock_filter.tear_down_filter.assert_called_once_with()
        self.mock__shutting_down.release.assert_called_once_with()

    def test_del_host_no_worker(self):
        self.manager._periodics_worker = None

        self.manager.del_host()

        self.mock__shutting_down.acquire.assert_called_once_with(
            blocking=False)
        self.mock__periodic_worker.stop.assert_not_called()
        self.mock__periodic_worker.wait.assert_not_called()
        self.assertIsNone(self.manager._periodics_worker)
        self.mock_executor.shutdown.assert_called_once_with(wait=True)
        self.mock_filter.tear_down_filter.assert_called_once_with()
        self.mock__shutting_down.release.assert_called_once_with()

    def test_del_host_stopped_executor(self):
        self.mock_executor.alive = False

        self.manager.del_host()

        self.mock__shutting_down.acquire.assert_called_once_with(
            blocking=False)
        self.mock__periodic_worker.stop.assert_called_once_with()
        self.mock__periodic_worker.wait.assert_called_once_with()
        self.assertIsNone(self.manager._periodics_worker)
        self.mock_executor.shutdown.assert_not_called()
        self.mock_filter.tear_down_filter.assert_called_once_with()
        self.mock__shutting_down.release.assert_called_once_with()


class TestManagerPeriodicWatchDog(BaseManagerTest):
    def setUp(self):
        super(TestManagerPeriodicWatchDog, self).setUp()
        self.mock_get_callable_name = self.useFixture(fixtures.MockPatchObject(
            manager.reflection, 'get_callable_name')).mock
        self.mock_spawn = self.useFixture(fixtures.MockPatchObject(
            manager.eventlet, 'spawn')).mock

    def test__periodics_watchdog(self):
        error = RuntimeError('Oops!')

        self.manager._periodics_watchdog(
            callable_=None, activity=None, spacing=None,
            exc_info=(None, error, None), traceback=None)

        self.mock_get_callable_name.assert_called_once_with(None)
        self.mock_spawn.assert_called_once_with(self.manager.del_host)


class TestManagerIntrospect(BaseManagerTest):
    @mock.patch.object(introspect, 'introspect', autospec=True)
    def test_do_introspect(self, introspect_mock):
        self.manager.do_introspection(self.context, self.uuid, self.token)

        introspect_mock.assert_called_once_with(self.uuid, token=self.token,
                                                manage_boot=True)

    @mock.patch.object(introspect, 'introspect', autospec=True)
    def test_do_introspect_with_manage_boot(self, introspect_mock):
        self.manager.do_introspection(self.context, self.uuid, self.token,
                                      False)

        introspect_mock.assert_called_once_with(self.uuid, token=self.token,
                                                manage_boot=False)

    @mock.patch.object(introspect, 'introspect', autospec=True)
    def test_introspect_failed(self, introspect_mock):
        introspect_mock.side_effect = utils.Error("boom")

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_introspection,
                                self.context, self.uuid, self.token)

        self.assertEqual(utils.Error, exc.exc_info[0])
        introspect_mock.assert_called_once_with(self.uuid, token=None,
                                                manage_boot=True)


class TestManagerAbort(BaseManagerTest):
    @mock.patch.object(introspect, 'abort', autospec=True)
    def test_abort_ok(self, abort_mock):
        self.manager.do_abort(self.context, self.uuid, self.token)

        abort_mock.assert_called_once_with(self.uuid, token=self.token)

    @mock.patch.object(introspect, 'abort', autospec=True)
    def test_abort_node_not_found(self, abort_mock):
        abort_mock.side_effect = utils.Error("Not Found.", code=404)

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_abort,
                                self.context, self.uuid, self.token)

        self.assertEqual(utils.Error, exc.exc_info[0])
        abort_mock.assert_called_once_with(self.uuid, token=None)

    @mock.patch.object(introspect, 'abort', autospec=True)
    def test_abort_failed(self, abort_mock):
        exc = utils.Error("Locked.", code=409)
        abort_mock.side_effect = exc

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_abort,
                                self.context, self.uuid, self.token)

        self.assertEqual(utils.Error, exc.exc_info[0])
        abort_mock.assert_called_once_with(self.uuid, token=None)


@mock.patch.object(process, 'reapply', autospec=True)
class TestManagerReapply(BaseManagerTest):

    def setUp(self):
        super(TestManagerReapply, self).setUp()
        CONF.set_override('store_data', 'swift', 'processing')

    @mock.patch.object(swift, 'store_introspection_data', autospec=True)
    @mock.patch.object(swift, 'get_introspection_data', autospec=True)
    def test_ok(self, swift_get_mock, swift_set_mock, reapply_mock):
        swift_get_mock.return_value = json.dumps(self.data)
        self.manager.do_reapply(self.context, self.uuid)
        reapply_mock.assert_called_once_with(self.uuid, data=self.data)

    @mock.patch.object(swift, 'store_introspection_data', autospec=True)
    @mock.patch.object(swift, 'get_introspection_data', autospec=True)
    def test_node_locked(self, swift_get_mock, swift_set_mock, reapply_mock):
        swift_get_mock.return_value = json.dumps(self.data)
        exc = utils.Error('Locked.', code=409)
        reapply_mock.side_effect = exc

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_reapply,
                                self.context, self.uuid)

        self.assertEqual(utils.Error, exc.exc_info[0])
        self.assertIn('Locked.', str(exc.exc_info[1]))
        self.assertEqual(409, exc.exc_info[1].http_code)
        reapply_mock.assert_called_once_with(self.uuid, data=self.data)

    @mock.patch.object(swift, 'store_introspection_data', autospec=True)
    @mock.patch.object(swift, 'get_introspection_data', autospec=True)
    def test_node_not_found(self, swift_get_mock, swift_set_mock,
                            reapply_mock):
        swift_get_mock.return_value = json.dumps(self.data)
        exc = utils.Error('Not found.', code=404)
        reapply_mock.side_effect = exc

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_reapply,
                                self.context, self.uuid)

        self.assertEqual(utils.Error, exc.exc_info[0])
        self.assertIn('Not found.', str(exc.exc_info[1]))
        self.assertEqual(404, exc.exc_info[1].http_code)
        reapply_mock.assert_called_once_with(self.uuid, data=self.data)

    @mock.patch.object(process, 'get_introspection_data', autospec=True)
    def test_generic_error(self, get_data_mock, reapply_mock):
        get_data_mock.return_value = self.data
        exc = utils.Error('Oops', code=400)
        reapply_mock.side_effect = exc

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_reapply,
                                self.context, self.uuid)

        self.assertEqual(utils.Error, exc.exc_info[0])
        self.assertIn('Oops', str(exc.exc_info[1]))
        self.assertEqual(400, exc.exc_info[1].http_code)
        reapply_mock.assert_called_once_with(self.uuid, data=self.data)
        get_data_mock.assert_called_once_with(self.uuid, processed=False,
                                              get_json=True)

    @mock.patch.object(process, 'get_introspection_data', autospec=True)
    def test_get_introspection_data_error(self, get_data_mock, reapply_mock):
        exc = utils.Error('The store is empty', code=404)
        get_data_mock.side_effect = exc

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_reapply,
                                self.context, self.uuid)

        self.assertEqual(utils.Error, exc.exc_info[0])
        self.assertIn('The store is empty', str(exc.exc_info[1]))
        self.assertEqual(404, exc.exc_info[1].http_code)
        get_data_mock.assert_called_once_with(self.uuid, processed=False,
                                              get_json=True)
        self.assertFalse(reapply_mock.called)

    def test_store_data_disabled(self, reapply_mock):
        CONF.set_override('store_data', 'none', 'processing')

        exc = self.assertRaises(messaging.rpc.ExpectedException,
                                self.manager.do_reapply,
                                self.context, self.uuid)

        self.assertEqual(utils.Error, exc.exc_info[0])
        self.assertIn('Inspector is not configured to store data',
                      str(exc.exc_info[1]))
        self.assertEqual(400, exc.exc_info[1].http_code)
        self.assertFalse(reapply_mock.called)

    @mock.patch.object(process, 'get_introspection_data', autospec=True)
    def test_ok_swift(self, get_data_mock, reapply_mock):
        get_data_mock.return_value = self.data
        self.manager.do_reapply(self.context, self.uuid)
        reapply_mock.assert_called_once_with(self.uuid, data=self.data)
        get_data_mock.assert_called_once_with(self.uuid, processed=False,
                                              get_json=True)

    @mock.patch.object(process, 'get_introspection_data', autospec=True)
    def test_ok_db(self, get_data_mock, reapply_mock):
        get_data_mock.return_value = self.data
        CONF.set_override('store_data', 'database', 'processing')
        self.manager.do_reapply(self.context, self.uuid)
        reapply_mock.assert_called_once_with(self.uuid, data=self.data)
        get_data_mock.assert_called_once_with(self.uuid, processed=False,
                                              get_json=True)



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Support for introspection rules."""

import jsonpath_rw as jsonpath
import jsonschema
from oslo_db import exception as db_exc
from oslo_utils import timeutils
from oslo_utils import uuidutils
import six
from sqlalchemy import orm

from ironic_inspector.common.i18n import _
from ironic_inspector import db
from ironic_inspector.plugins import base as plugins_base
from ironic_inspector import utils


LOG = utils.getProcessingLogger(__name__)
_CONDITIONS_SCHEMA = None
_ACTIONS_SCHEMA = None


def conditions_schema():
    global _CONDITIONS_SCHEMA
    if _CONDITIONS_SCHEMA is None:
        condition_plugins = [x.name for x in
                             plugins_base.rule_conditions_manager()]
        _CONDITIONS_SCHEMA = {
            "title": "Inspector rule conditions schema",
            "type": "array",
            # we can have rules that always apply
            "minItems": 0,
            "items": {
                "type": "object",
                # field might become optional in the future, but not right now
                "required": ["op", "field"],
                "properties": {
                    "op": {
                        "description": "condition operator",
                        "enum": condition_plugins
                    },
                    "field": {
                        "description": "JSON path to field for matching",
                        "type": "string"
                    },
                    "multiple": {
                        "description": "how to treat multiple values",
                        "enum": ["all", "any", "first"]
                    },
                    "invert": {
                        "description": "whether to invert the result",
                        "type": "boolean"
                    },
                },
                # other properties are validated by plugins
                "additionalProperties": True
            }
        }

    return _CONDITIONS_SCHEMA


def actions_schema():
    global _ACTIONS_SCHEMA
    if _ACTIONS_SCHEMA is None:
        action_plugins = [x.name for x in
                          plugins_base.rule_actions_manager()]
        _ACTIONS_SCHEMA = {
            "title": "Inspector rule actions schema",
            "type": "array",
            "minItems": 1,
            "items": {
                "type": "object",
                "required": ["action"],
                "properties": {
                    "action": {
                        "description": "action to take",
                        "enum": action_plugins
                    },
                },
                # other properties are validated by plugins
                "additionalProperties": True
            }
        }

    return _ACTIONS_SCHEMA


class IntrospectionRule(object):
    """High-level class representing an introspection rule."""

    def __init__(self, uuid, conditions, actions, description):
        """Create rule object from database data."""
        self._uuid = uuid
        self._conditions = conditions
        self._actions = actions
        self._description = description

    def as_dict(self, short=False):
        result = {
            'uuid': self._uuid,
            'description': self._description,
        }

        if not short:
            result['conditions'] = [c.as_dict() for c in self._conditions]
            result['actions'] = [a.as_dict() for a in self._actions]

        return result

    @property
    def description(self):
        return self._description or self._uuid

    def check_conditions(self, node_info, data):
        """Check if conditions are true for a given node.

        :param node_info: a NodeInfo object
        :param data: introspection data
        :returns: True if conditions match, otherwise False
        """
        LOG.debug('Checking rule "%s"', self.description,
                  node_info=node_info, data=data)
        ext_mgr = plugins_base.rule_conditions_manager()
        for cond in self._conditions:
            scheme, path = _parse_path(cond.field)

            if scheme == 'node':
                source_data = node_info.node().to_dict()
            elif scheme == 'data':
                source_data = data

            field_values = jsonpath.parse(path).find(source_data)
            field_values = [x.value for x in field_values]
            cond_ext = ext_mgr[cond.op].obj

            if not field_values:
                if cond_ext.ALLOW_NONE:
                    LOG.debug('Field with JSON path %s was not found in data',
                              cond.field, node_info=node_info, data=data)
                    field_values = [None]
                else:
                    LOG.info('Field with JSON path %(path)s was not found '
                             'in data, rule "%(rule)s" will not '
                             'be applied',
                             {'path': cond.field, 'rule': self.description},
                             node_info=node_info, data=data)
                    return False

            for value in field_values:
                result = cond_ext.check(node_info, value, cond.params)
                if cond.invert:
                    result = not result

                if (cond.multiple == 'first'
                        or (cond.multiple == 'all' and not result)
                        or (cond.multiple == 'any' and result)):
                    break

            if not result:
                LOG.info('Rule "%(rule)s" will not be applied: condition '
                         '%(field)s %(op)s %(params)s failed',
                         {'rule': self.description, 'field': cond.field,
                          'op': cond.op, 'params': cond.params},
                         node_info=node_info, data=data)
                return False

        LOG.info('Rule "%s" will be applied', self.description,
                 node_info=node_info, data=data)
        return True

    def apply_actions(self, node_info, data=None):
        """Run actions on a node.

        :param node_info: NodeInfo instance
        :param data: introspection data
        """
        LOG.debug('Running actions for rule "%s"', self.description,
                  node_info=node_info, data=data)

        ext_mgr = plugins_base.rule_actions_manager()
        for act in self._actions:
            ext = ext_mgr[act.action].obj

            for formatted_param in ext.FORMATTED_PARAMS:
                try:
                    initial = act.params[formatted_param]
                except KeyError:
                    # Ignore parameter that wasn't given.
                    continue
                else:
                    act.params[formatted_param] = _format_value(initial, data)

            LOG.debug('Running action `%(action)s %(params)s`',
                      {'action': act.action, 'params': act.params},
                      node_info=node_info, data=data)
            ext.apply(node_info, act.params)

        LOG.debug('Successfully applied actions',
                  node_info=node_info, data=data)


def _format_value(value, data):
    """Apply parameter formatting to a value.

    Format strings with the values from `data`. If `value` is a dict or
    list, any string members (and any nested string members) will also be
    formatted recursively. This includes both keys and values for dicts.

    :param value: The string to format, or container whose members to
                  format.
    :param data: Introspection data.
    :returns: `value`, formatted with the parameters from `data`.
    """
    if isinstance(value, six.string_types):
        # NOTE(aarefiev): verify provided value with introspection
        # data format specifications.
        # TODO(aarefiev): simple verify on import rule time.
        try:
            return value.format(data=data)
        except KeyError as e:
            raise utils.Error(_('Invalid formatting variable key '
                                'provided in value %(val)s: %(e)s'),
                              {'val': value, 'e': e}, data=data)
    elif isinstance(value, dict):
        return {_format_value(k, data): _format_value(v, data)
                for k, v in six.iteritems(value)}
    elif isinstance(value, list):
        return [_format_value(v, data) for v in value]
    else:
        # Assume this is a 'primitive' value.
        return value


def _parse_path(path):
    """Parse path, extract scheme and path.

     Parse path with 'node' and 'data' scheme, which links on
     introspection data and node info respectively. If scheme is
     missing in path, default is 'data'.

    :param path: data or node path
    :return: tuple (scheme, path)
    """
    try:
        index = path.index('://')
    except ValueError:
        scheme = 'data'
        path = path
    else:
        scheme = path[:index]
        path = path[index + 3:]
    return scheme, path


def _validate_conditions(conditions_json):
    """Validates conditions from jsonschema.

    :returns: a list of conditions.
    """
    try:
        jsonschema.validate(conditions_json, conditions_schema())
    except jsonschema.ValidationError as exc:
        raise utils.Error(_('Validation failed for conditions: %s') % exc)

    cond_mgr = plugins_base.rule_conditions_manager()
    conditions = []
    reserved_params = {'op', 'field', 'multiple', 'invert'}
    for cond_json in conditions_json:
        field = cond_json['field']

        scheme, path = _parse_path(field)

        if scheme not in ('node', 'data'):
            raise utils.Error(_('Unsupported scheme for field: %s, valid '
                                'values are node:// or data://') % scheme)
        # verify field as JSON path
        try:
            jsonpath.parse(path)
        except Exception as exc:
            raise utils.Error(_('Unable to parse field JSON path %(field)s: '
                                '%(error)s') % {'field': field, 'error': exc})

        plugin = cond_mgr[cond_json['op']].obj
        params = {k: v for k, v in cond_json.items()
                  if k not in reserved_params}
        try:
            plugin.validate(params)
        except ValueError as exc:
            raise utils.Error(_('Invalid parameters for operator %(op)s: '
                                '%(error)s') %
                              {'op': cond_json['op'], 'error': exc})

        conditions.append((cond_json['field'],
                           cond_json['op'],
                           cond_json.get('multiple', 'any'),
                           cond_json.get('invert', False),
                           params))
    return conditions


def _validate_actions(actions_json):
    """Validates actions from jsonschema.

    :returns: a list of actions.
    """
    try:
        jsonschema.validate(actions_json, actions_schema())
    except jsonschema.ValidationError as exc:
        raise utils.Error(_('Validation failed for actions: %s') % exc)

    act_mgr = plugins_base.rule_actions_manager()
    actions = []
    for action_json in actions_json:
        plugin = act_mgr[action_json['action']].obj
        params = {k: v for k, v in action_json.items() if k != 'action'}
        try:
            plugin.validate(params)
        except ValueError as exc:
            raise utils.Error(_('Invalid parameters for action %(act)s: '
                                '%(error)s') %
                              {'act': action_json['action'], 'error': exc})

        actions.append((action_json['action'], params))
    return actions


def create(conditions_json, actions_json, uuid=None,
           description=None):
    """Create a new rule in database.

    :param conditions_json: list of dicts with the following keys:
                            * op - operator
                            * field - JSON path to field to compare
                            Other keys are stored as is.
    :param actions_json: list of dicts with the following keys:
                         * action - action type
                         Other keys are stored as is.
    :param uuid: rule UUID, will be generated if empty
    :param description: human-readable rule description
    :returns: new IntrospectionRule object
    :raises: utils.Error on failure
    """
    uuid = uuid or uuidutils.generate_uuid()
    LOG.debug('Creating rule %(uuid)s with description "%(descr)s", '
              'conditions %(conditions)s and actions %(actions)s',
              {'uuid': uuid, 'descr': description,
               'conditions': conditions_json, 'actions': actions_json})

    conditions = _validate_conditions(conditions_json)
    actions = _validate_actions(actions_json)

    try:
        with db.ensure_transaction() as session:
            rule = db.Rule(uuid=uuid, description=description,
                           disabled=False, created_at=timeutils.utcnow())

            for field, op, multiple, invert, params in conditions:
                rule.conditions.append(db.RuleCondition(op=op,
                                                        field=field,
                                                        multiple=multiple,
                                                        invert=invert,
                                                        params=params))

            for action, params in actions:
                rule.actions.append(db.RuleAction(action=action,
                                                  params=params))

            rule.save(session)
    except db_exc.DBDuplicateEntry as exc:
        LOG.error('Database integrity error %s when '
                  'creating a rule', exc)
        raise utils.Error(_('Rule with UUID %s already exists') % uuid,
                          code=409)

    LOG.info('Created rule %(uuid)s with description "%(descr)s"',
             {'uuid': uuid, 'descr': description})
    return IntrospectionRule(uuid=uuid,
                             conditions=rule.conditions,
                             actions=rule.actions,
                             description=description)


def get(uuid):
    """Get a rule by its UUID."""
    try:
        rule = db.model_query(db.Rule).filter_by(uuid=uuid).one()
    except orm.exc.NoResultFound:
        raise utils.Error(_('Rule %s was not found') % uuid, code=404)

    return IntrospectionRule(uuid=rule.uuid, actions=rule.actions,
                             conditions=rule.conditions,
                             description=rule.description)


def get_all():
    """List all rules."""
    query = db.model_query(db.Rule).order_by(db.Rule.created_at)
    return [IntrospectionRule(uuid=rule.uuid, actions=rule.actions,
                              conditions=rule.conditions,
                              description=rule.description)
            for rule in query]


def delete(uuid):
    """Delete a rule by its UUID."""
    with db.ensure_transaction() as session:
        db.model_query(db.RuleAction,
                       session=session).filter_by(rule=uuid).delete()
        db.model_query(db.RuleCondition,
                       session=session) .filter_by(rule=uuid).delete()
        count = (db.model_query(db.Rule, session=session)
                 .filter_by(uuid=uuid).delete())
        if not count:
            raise utils.Error(_('Rule %s was not found') % uuid, code=404)

    LOG.info('Introspection rule %s was deleted', uuid)


def delete_all():
    """Delete all rules."""
    with db.ensure_transaction() as session:
        db.model_query(db.RuleAction, session=session).delete()
        db.model_query(db.RuleCondition, session=session).delete()
        db.model_query(db.Rule, session=session).delete()

    LOG.info('All introspection rules were deleted')


def apply(node_info, data):
    """Apply rules to a node."""
    rules = get_all()
    if not rules:
        LOG.debug('No custom introspection rules to apply',
                  node_info=node_info, data=data)
        return

    LOG.debug('Applying custom introspection rules',
              node_info=node_info, data=data)

    to_apply = []
    for rule in rules:
        if rule.check_conditions(node_info, data):
            to_apply.append(rule)

    if to_apply:
        LOG.debug('Running actions', node_info=node_info, data=data)
        for rule in to_apply:
            rule.apply_actions(node_info, data=data)
    else:
        LOG.debug('No actions to apply', node_info=node_info, data=data)

    LOG.info('Successfully applied custom introspection rules',
             node_info=node_info, data=data)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from automaton import exceptions as automaton_errors
from eventlet import semaphore
import fixtures
from futurist import periodics
import mock
from oslo_config import cfg
import six
import stevedore

from ironic_inspector.common import ironic as ir_utils
from ironic_inspector.pxe_filter import base as pxe_filter
from ironic_inspector.pxe_filter import interface
from ironic_inspector.test import base as test_base

CONF = cfg.CONF


class TestDriverManager(test_base.BaseTest):
    def setUp(self):
        super(TestDriverManager, self).setUp()
        pxe_filter._DRIVER_MANAGER = None
        stevedore_driver_fixture = self.useFixture(fixtures.MockPatchObject(
            stevedore.driver, 'DriverManager', autospec=True))
        self.stevedore_driver_mock = stevedore_driver_fixture.mock

    def test_default(self):
        driver_manager = pxe_filter._driver_manager()
        self.stevedore_driver_mock.assert_called_once_with(
            pxe_filter._STEVEDORE_DRIVER_NAMESPACE,
            name='iptables',
            invoke_on_load=True
        )
        self.assertIsNotNone(driver_manager)
        self.assertIs(pxe_filter._DRIVER_MANAGER, driver_manager)

    def test_pxe_filter_name(self):
        CONF.set_override('driver', 'foo', 'pxe_filter')
        driver_manager = pxe_filter._driver_manager()
        self.stevedore_driver_mock.assert_called_once_with(
            pxe_filter._STEVEDORE_DRIVER_NAMESPACE,
            'foo',
            invoke_on_load=True
        )
        self.assertIsNotNone(driver_manager)
        self.assertIs(pxe_filter._DRIVER_MANAGER, driver_manager)

    def test_default_existing_driver_manager(self):
        pxe_filter._DRIVER_MANAGER = True
        driver_manager = pxe_filter._driver_manager()
        self.stevedore_driver_mock.assert_not_called()
        self.assertIs(pxe_filter._DRIVER_MANAGER, driver_manager)


class TestDriverManagerLoading(test_base.BaseTest):
    def setUp(self):
        super(TestDriverManagerLoading, self).setUp()
        pxe_filter._DRIVER_MANAGER = None

    @mock.patch.object(pxe_filter, 'NoopFilter', autospec=True)
    def test_pxe_filter_driver_loads(self, noop_driver_cls):
        CONF.set_override('driver', 'noop', 'pxe_filter')
        driver_manager = pxe_filter._driver_manager()
        noop_driver_cls.assert_called_once_with()
        self.assertIs(noop_driver_cls.return_value, driver_manager.driver)

    def test_invalid_filter_driver(self):
        CONF.set_override('driver', 'foo', 'pxe_filter')
        six.assertRaisesRegex(self, stevedore.exception.NoMatches, 'foo',
                              pxe_filter._driver_manager)
        self.assertIsNone(pxe_filter._DRIVER_MANAGER)


class BaseFilterBaseTest(test_base.BaseTest):
    def setUp(self):
        super(BaseFilterBaseTest, self).setUp()
        self.mock_lock = mock.MagicMock(spec=semaphore.BoundedSemaphore)
        self.mock_bounded_semaphore = self.useFixture(
            fixtures.MockPatchObject(semaphore, 'BoundedSemaphore')).mock
        self.mock_bounded_semaphore.return_value = self.mock_lock
        self.driver = pxe_filter.NoopFilter()

    def assert_driver_is_locked(self):
        """Assert the driver is currently locked and wasn't locked before."""
        self.driver.lock.__enter__.assert_called_once_with()
        self.driver.lock.__exit__.assert_not_called()

    def assert_driver_was_locked_once(self):
        """Assert the driver was locked exactly once before."""
        self.driver.lock.__enter__.assert_called_once_with()
        self.driver.lock.__exit__.assert_called_once_with(None, None, None)

    def assert_driver_was_not_locked(self):
        """Assert the driver was not locked"""
        self.mock_lock.__enter__.assert_not_called()
        self.mock_lock.__exit__.assert_not_called()


class TestLockedDriverEvent(BaseFilterBaseTest):
    def setUp(self):
        super(TestLockedDriverEvent, self).setUp()
        self.mock_fsm_reset_on_error = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'fsm_reset_on_error')).mock
        self.expected_args = (None,)
        self.expected_kwargs = {'foo': None}
        self.mock_fsm = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'fsm')).mock
        (self.driver.fsm_reset_on_error.return_value.
            __enter__.return_value) = self.mock_fsm

    def test_locked_driver_event(self):
        event = 'foo'

        @pxe_filter.locked_driver_event(event)
        def fun(driver, *args, **kwargs):
            self.assertIs(self.driver, driver)
            self.assertEqual(self.expected_args, args)
            self.assertEqual(self.expected_kwargs, kwargs)
            self.assert_driver_is_locked()

        self.assert_driver_was_not_locked()
        fun(self.driver, *self.expected_args, **self.expected_kwargs)

        self.mock_fsm_reset_on_error.assert_called_once_with()
        self.mock_fsm.process_event.assert_called_once_with(event)
        self.assert_driver_was_locked_once()


class TestBaseFilterFsmPrecautions(BaseFilterBaseTest):
    def setUp(self):
        super(TestBaseFilterFsmPrecautions, self).setUp()
        self.mock_fsm = self.useFixture(
            fixtures.MockPatchObject(pxe_filter.NoopFilter, 'fsm')).mock
        # NOTE(milan): overriding driver so that the patch ^ is applied
        self.mock_bounded_semaphore.reset_mock()
        self.driver = pxe_filter.NoopFilter()
        self.mock_reset = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'reset')).mock

    def test___init__(self):
        self.assertIs(self.mock_lock, self.driver.lock)
        self.mock_bounded_semaphore.assert_called_once_with()
        self.assertIs(self.mock_fsm, self.driver.fsm)
        self.mock_fsm.initialize.assert_called_once_with(
            start_state=pxe_filter.States.uninitialized)

    def test_fsm_reset_on_error(self):
        with self.driver.fsm_reset_on_error() as fsm:
            self.assertIs(self.mock_fsm, fsm)

        self.mock_reset.assert_not_called()

    def test_fsm_automaton_error(self):

        def fun():
            with self.driver.fsm_reset_on_error():
                raise automaton_errors.NotFound('Oops!')

        self.assertRaisesRegex(pxe_filter.InvalidFilterDriverState,
                               '.*NoopFilter.*Oops!', fun)
        self.mock_reset.assert_not_called()

    def test_fsm_reset_on_error_ctx_custom_error(self):

        class MyError(Exception):
            pass

        def fun():
            with self.driver.fsm_reset_on_error():
                raise MyError('Oops!')

        self.assertRaisesRegex(MyError, 'Oops!', fun)
        self.mock_reset.assert_called_once_with()


class TestBaseFilterInterface(BaseFilterBaseTest):
    def setUp(self):
        super(TestBaseFilterInterface, self).setUp()
        self.mock_get_client = self.useFixture(
            fixtures.MockPatchObject(ir_utils, 'get_client')).mock
        self.mock_ironic = mock.Mock()
        self.mock_get_client.return_value = self.mock_ironic
        self.mock_periodic = self.useFixture(
            fixtures.MockPatchObject(periodics, 'periodic')).mock
        self.mock_reset = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'reset')).mock
        self.mock_log = self.useFixture(
            fixtures.MockPatchObject(pxe_filter, 'LOG')).mock
        self.driver.fsm_reset_on_error = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'fsm_reset_on_error')).mock

    def test_init_filter(self):
        self.driver.init_filter()

        self.mock_log.debug.assert_called_once_with(
            'Initializing the PXE filter driver %s', self.driver)
        self.mock_reset.assert_not_called()

    def test_sync(self):
        self.driver.sync(self.mock_ironic)

        self.mock_reset.assert_not_called()

    def test_tear_down_filter(self):
        self.assert_driver_was_not_locked()
        self.driver.tear_down_filter()

        self.assert_driver_was_locked_once()
        self.mock_reset.assert_called_once_with()

    def test_get_periodic_sync_task(self):
        sync_mock = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'sync')).mock
        self.driver.get_periodic_sync_task()
        self.mock_periodic.assert_called_once_with(spacing=15, enabled=True)
        self.mock_periodic.return_value.call_args[0][0]()
        sync_mock.assert_called_once_with(self.mock_get_client.return_value)

    def test_get_periodic_sync_task_invalid_state(self):
        sync_mock = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'sync')).mock
        sync_mock.side_effect = pxe_filter.InvalidFilterDriverState('Oops!')

        self.driver.get_periodic_sync_task()
        self.mock_periodic.assert_called_once_with(spacing=15, enabled=True)
        self.assertRaisesRegex(periodics.NeverAgain, 'Oops!',
                               self.mock_periodic.return_value.call_args[0][0])

    def test_get_periodic_sync_task_custom_error(self):
        class MyError(Exception):
            pass

        sync_mock = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'sync')).mock
        sync_mock.side_effect = MyError('Oops!')

        self.driver.get_periodic_sync_task()
        self.mock_periodic.assert_called_once_with(spacing=15, enabled=True)
        self.assertRaisesRegex(
            MyError, 'Oops!', self.mock_periodic.return_value.call_args[0][0])

    def test_get_periodic_sync_task_disabled(self):
        CONF.set_override('sync_period', 0, 'pxe_filter')
        self.driver.get_periodic_sync_task()
        self.mock_periodic.assert_called_once_with(spacing=float('inf'),
                                                   enabled=False)

    def test_get_periodic_sync_task_custom_spacing(self):
        CONF.set_override('sync_period', 4224, 'pxe_filter')
        self.driver.get_periodic_sync_task()
        self.mock_periodic.assert_called_once_with(spacing=4224, enabled=True)


class TestDriverReset(BaseFilterBaseTest):
    def setUp(self):
        super(TestDriverReset, self).setUp()
        self.mock_fsm = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'fsm')).mock

    def test_reset(self):
        self.driver.reset()

        self.assert_driver_was_not_locked()
        self.mock_fsm.process_event.assert_called_once_with(
            pxe_filter.Events.reset)


class TestDriver(test_base.BaseTest):
    def setUp(self):
        super(TestDriver, self).setUp()
        self.mock_driver = mock.Mock(spec=interface.FilterDriver)
        self.mock__driver_manager = self.useFixture(
            fixtures.MockPatchObject(pxe_filter, '_driver_manager')).mock
        self.mock__driver_manager.return_value.driver = self.mock_driver

    def test_driver(self):
        ret = pxe_filter.driver()

        self.assertIs(self.mock_driver, ret)
        self.mock__driver_manager.assert_called_once_with()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Mostly copied from ironic/common/swift.py

import json

from oslo_config import cfg
from swiftclient import client as swift_client
from swiftclient import exceptions as swift_exceptions

from ironic_inspector.common.i18n import _
from ironic_inspector.common import keystone
from ironic_inspector import utils

CONF = cfg.CONF


OBJECT_NAME_PREFIX = 'inspector_data'
SWIFT_SESSION = None


def reset_swift_session():
    """Reset the global session variable.

    Mostly useful for unit tests.
    """
    global SWIFT_SESSION
    SWIFT_SESSION = None


class SwiftAPI(object):
    """API for communicating with Swift."""

    def __init__(self):
        """Constructor for creating a SwiftAPI object.

        Authentification is loaded from config file.
        """
        global SWIFT_SESSION
        if not SWIFT_SESSION:
            SWIFT_SESSION = keystone.get_session('swift')

        adapter_opts = dict()
        # TODO(pas-ha): remove handling deprecated options in Rocky
        if CONF.swift.os_region and not CONF.swift.region_name:
            adapter_opts['region_name'] = CONF.swift.os_region

        adapter = keystone.get_adapter('swift', session=SWIFT_SESSION,
                                       **adapter_opts)

        # TODO(pas-ha) reverse-construct SSL-related session options here
        params = {
            'os_options': {
                'object_storage_url': adapter.get_endpoint()}}

        self.connection = swift_client.Connection(session=SWIFT_SESSION,
                                                  **params)

    def create_object(self, object, data, container=CONF.swift.container,
                      headers=None):
        """Uploads a given string to Swift.

        :param object: The name of the object in Swift
        :param data: string data to put in the object
        :param container: The name of the container for the object.
        :param headers: the headers for the object to pass to Swift
        :returns: The Swift UUID of the object
        :raises: utils.Error, if any operation with Swift fails.
        """
        try:
            self.connection.put_container(container)
        except swift_exceptions.ClientException as e:
            err_msg = (_('Swift failed to create container %(container)s. '
                         'Error was: %(error)s') %
                       {'container': container, 'error': e})
            raise utils.Error(err_msg)

        if CONF.swift.delete_after > 0:
            headers = headers or {}
            headers['X-Delete-After'] = CONF.swift.delete_after

        try:
            obj_uuid = self.connection.put_object(container,
                                                  object,
                                                  data,
                                                  headers=headers)
        except swift_exceptions.ClientException as e:
            err_msg = (_('Swift failed to create object %(object)s in '
                         'container %(container)s. Error was: %(error)s') %
                       {'object': object, 'container': container, 'error': e})
            raise utils.Error(err_msg)

        return obj_uuid

    def get_object(self, object, container=CONF.swift.container):
        """Downloads a given object from Swift.

        :param object: The name of the object in Swift
        :param container: The name of the container for the object.
        :returns: Swift object
        :raises: utils.Error, if the Swift operation fails.
        """
        try:
            headers, obj = self.connection.get_object(container, object)
        except swift_exceptions.ClientException as e:
            err_msg = (_('Swift failed to get object %(object)s in '
                         'container %(container)s. Error was: %(error)s') %
                       {'object': object, 'container': container, 'error': e})
            raise utils.Error(err_msg)

        return obj


def store_introspection_data(data, uuid, suffix=None):
    """Uploads introspection data to Swift.

    :param data: data to store in Swift
    :param uuid: UUID of the Ironic node that the data came from
    :param suffix: optional suffix to add to the underlying swift
                   object name
    :returns: name of the Swift object that the data is stored in
    """
    swift_api = SwiftAPI()
    swift_object_name = '%s-%s' % (OBJECT_NAME_PREFIX, uuid)
    if suffix is not None:
        swift_object_name = '%s-%s' % (swift_object_name, suffix)
    swift_api.create_object(swift_object_name, json.dumps(data))
    return swift_object_name


def get_introspection_data(uuid, suffix=None):
    """Downloads introspection data from Swift.

    :param uuid: UUID of the Ironic node that the data came from
    :param suffix: optional suffix to add to the underlying swift
                   object name
    :returns: Swift object with the introspection data
    """
    swift_api = SwiftAPI()
    swift_object_name = '%s-%s' % (OBJECT_NAME_PREFIX, uuid)
    if suffix is not None:
        swift_object_name = '%s-%s' % (swift_object_name, suffix)
    return swift_api.get_object(swift_object_name)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import mock

from ironic_inspector import node_cache
from ironic_inspector.plugins import extra_hardware
from ironic_inspector.test import base as test_base


@mock.patch.object(extra_hardware.swift, 'SwiftAPI', autospec=True)
@mock.patch.object(node_cache.NodeInfo, 'patch')
class TestExtraHardware(test_base.NodeTest):
    hook = extra_hardware.ExtraHardwareHook()

    def test_data_recieved(self, patch_mock, swift_mock):
        introspection_data = {
            'data': [['memory', 'total', 'size', '4294967296'],
                     ['cpu', 'physical', 'number', '1'],
                     ['cpu', 'logical', 'number', '1']]}
        data = json.dumps(introspection_data['data'])
        self.hook.before_processing(introspection_data)
        self.hook.before_update(introspection_data, self.node_info)

        swift_conn = swift_mock.return_value
        name = 'extra_hardware-%s' % self.uuid
        swift_conn.create_object.assert_called_once_with(name, data)
        patch_mock.assert_called_once_with(
            [{'op': 'add', 'path': '/extra/hardware_swift_object',
              'value': name}])

        expected = {
            'memory': {
                'total': {
                    'size': 4294967296
                }
            },
            'cpu': {
                'physical': {
                    'number': 1
                },
                'logical': {
                    'number': 1
                },
            }
        }

        self.assertEqual(expected, introspection_data['extra'])

    def test_data_not_in_edeploy_format(self, patch_mock, swift_mock):
        introspection_data = {
            'data': [['memory', 'total', 'size', '4294967296'],
                     ['cpu', 'physical', 'number', '1'],
                     {'interface': 'eth1'}]}
        data = json.dumps(introspection_data['data'])
        self.hook.before_processing(introspection_data)
        self.hook.before_update(introspection_data, self.node_info)

        swift_conn = swift_mock.return_value
        name = 'extra_hardware-%s' % self.uuid
        swift_conn.create_object.assert_called_once_with(name, data)
        patch_mock.assert_called_once_with(
            [{'op': 'add', 'path': '/extra/hardware_swift_object',
              'value': name}])

        self.assertNotIn('data', introspection_data)

    def test_no_data_recieved(self, patch_mock, swift_mock):
        introspection_data = {'cats': 'meow'}
        swift_conn = swift_mock.return_value
        self.hook.before_processing(introspection_data)
        self.hook.before_update(introspection_data, self.node_info)
        self.assertFalse(patch_mock.called)
        self.assertFalse(swift_conn.create_object.called)

    def test__convert_edeploy_data(self, patch_mock, swift_mock):
        introspection_data = [['Sheldon', 'J.', 'Plankton', '123'],
                              ['Larry', 'the', 'Lobster', None],
                              ['Eugene', 'H.', 'Krabs', 'The cashier']]

        data = self.hook._convert_edeploy_data(introspection_data)
        expected_data = {'Sheldon': {'J.': {'Plankton': 123}},
                         'Larry': {'the': {'Lobster': None}},
                         'Eugene': {'H.': {'Krabs': 'The cashier'}}}
        self.assertEqual(expected_data, data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from oslo_config import cfg

from oslo_config import cfg

from ironic_inspector.conf import capabilities
from ironic_inspector.conf import default
from ironic_inspector.conf import discovery
from ironic_inspector.conf import dnsmasq_pxe_filter
from ironic_inspector.conf import iptables
from ironic_inspector.conf import ironic
from ironic_inspector.conf import pci_devices
from ironic_inspector.conf import processing
from ironic_inspector.conf import pxe_filter
from ironic_inspector.conf import swift


CONF = cfg.CONF


capabilities.register_opts(CONF)
discovery.register_opts(CONF)
default.register_opts(CONF)
dnsmasq_pxe_filter.register_opts(CONF)
iptables.register_opts(CONF)
ironic.register_opts(CONF)
pci_devices.register_opts(CONF)
processing.register_opts(CONF)
pxe_filter.register_opts(CONF)
swift.register_opts(CONF)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Mostly copied from ironic/tests/test_swift.py


from keystoneauth1 import loading as kloading
import mock
from swiftclient import client as swift_client
from swiftclient import exceptions as swift_exception

from ironic_inspector.common import keystone
from ironic_inspector.common import swift
from ironic_inspector.test import base as test_base
from ironic_inspector import utils


class BaseTest(test_base.NodeTest):
    def setUp(self):
        super(BaseTest, self).setUp()
        self.all_macs = self.macs + ['DE:AD:BE:EF:DE:AD']
        self.pxe_mac = self.macs[1]
        self.data = {
            'ipmi_address': self.bmc_address,
            'cpus': 2,
            'cpu_arch': 'x86_64',
            'memory_mb': 1024,
            'local_gb': 20,
            'interfaces': {
                'em1': {'mac': self.macs[0], 'ip': '1.2.0.1'},
                'em2': {'mac': self.macs[1], 'ip': '1.2.0.2'},
                'em3': {'mac': self.all_macs[2]},
            },
            'boot_interface': '01-' + self.pxe_mac.replace(':', '-'),
        }


@mock.patch.object(keystone, 'get_adapter', autospec=True)
@mock.patch.object(keystone, 'register_auth_opts')
@mock.patch.object(keystone, 'get_session')
@mock.patch.object(swift_client, 'Connection', autospec=True)
class SwiftTestCase(BaseTest):

    def setUp(self):
        super(SwiftTestCase, self).setUp()
        swift.reset_swift_session()
        self.swift_exception = swift_exception.ClientException('', '')
        self.cfg.config(group='swift',
                        os_service_type='object-store',
                        os_endpoint_type='internalURL',
                        os_region='somewhere',
                        max_retries=2)
        # NOTE(aarefiev) register keystoneauth dynamic options
        adapter_opts = kloading.get_adapter_conf_options(
            include_deprecated=False)
        self.cfg.register_opts(adapter_opts, 'swift')
        self.addCleanup(swift.reset_swift_session)

    def test___init__(self, connection_mock, load_mock,
                      opts_mock, adapter_mock):
        fake_endpoint = "http://localhost:6000"
        adapter_mock.return_value.get_endpoint.return_value = fake_endpoint
        swift.SwiftAPI()
        connection_mock.assert_called_once_with(
            session=load_mock.return_value,
            os_options={'object_storage_url': fake_endpoint})

    def test_create_object(self, connection_mock, load_mock,
                           opts_mock, adapter_mock):
        swiftapi = swift.SwiftAPI()
        connection_obj_mock = connection_mock.return_value

        connection_obj_mock.put_object.return_value = 'object-uuid'

        object_uuid = swiftapi.create_object('object', 'some-string-data')

        connection_obj_mock.put_container.assert_called_once_with('ironic-'
                                                                  'inspector')
        connection_obj_mock.put_object.assert_called_once_with(
            'ironic-inspector', 'object', 'some-string-data', headers=None)
        self.assertEqual('object-uuid', object_uuid)

    def test_create_object_create_container_fails(
            self, connection_mock, load_mock, opts_mock, adapter_mock):
        swiftapi = swift.SwiftAPI()
        connection_obj_mock = connection_mock.return_value
        connection_obj_mock.put_container.side_effect = self.swift_exception
        self.assertRaises(utils.Error, swiftapi.create_object, 'object',
                          'some-string-data')
        connection_obj_mock.put_container.assert_called_once_with('ironic-'
                                                                  'inspector')
        self.assertFalse(connection_obj_mock.put_object.called)

    def test_create_object_put_object_fails(self, connection_mock, load_mock,
                                            opts_mock, adapter_mock):
        swiftapi = swift.SwiftAPI()
        connection_obj_mock = connection_mock.return_value
        connection_obj_mock.put_object.side_effect = self.swift_exception
        self.assertRaises(utils.Error, swiftapi.create_object, 'object',
                          'some-string-data')
        connection_obj_mock.put_container.assert_called_once_with('ironic-'
                                                                  'inspector')
        connection_obj_mock.put_object.assert_called_once_with(
            'ironic-inspector', 'object', 'some-string-data', headers=None)

    def test_get_object(self, connection_mock, load_mock,
                        opts_mock, adapter_mock):
        swiftapi = swift.SwiftAPI()
        connection_obj_mock = connection_mock.return_value

        expected_obj = self.data
        connection_obj_mock.get_object.return_value = ('headers', expected_obj)

        swift_obj = swiftapi.get_object('object')

        connection_obj_mock.get_object.assert_called_once_with(
            'ironic-inspector', 'object')
        self.assertEqual(expected_obj, swift_obj)

    def test_get_object_fails(self, connection_mock, load_mock,
                              opts_mock, adapter_mock):
        swiftapi = swift.SwiftAPI()
        connection_obj_mock = connection_mock.return_value
        connection_obj_mock.get_object.side_effect = self.swift_exception
        self.assertRaises(utils.Error, swiftapi.get_object,
                          'object')
        connection_obj_mock.get_object.assert_called_once_with(
            'ironic-inspector', 'object')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

import flask
from oslo_utils import strutils
from oslo_utils import uuidutils
import six
import werkzeug

from ironic_inspector import api_tools
from ironic_inspector.common import context
from ironic_inspector.common.i18n import _
from ironic_inspector.common import ironic as ir_utils
from ironic_inspector.common import rpc
import ironic_inspector.conf
from ironic_inspector.conf import opts as conf_opts
from ironic_inspector import node_cache
from ironic_inspector import process
from ironic_inspector import rules
from ironic_inspector import utils

CONF = ironic_inspector.conf.CONF


app = flask.Flask(__name__)
LOG = utils.getProcessingLogger(__name__)

MINIMUM_API_VERSION = (1, 0)
CURRENT_API_VERSION = (1, 14)
DEFAULT_API_VERSION = CURRENT_API_VERSION
_LOGGING_EXCLUDED_KEYS = ('logs',)


def _get_version():
    ver = flask.request.headers.get(conf_opts.VERSION_HEADER,
                                    _DEFAULT_API_VERSION)
    try:
        if ver.lower() == 'latest':
            requested = CURRENT_API_VERSION
        else:
            requested = tuple(int(x) for x in ver.split('.'))
    except (ValueError, TypeError):
        return error_response(_('Malformed API version: expected string '
                                'in form of X.Y or latest'), code=400)
    return requested


def _format_version(ver):
    return '%d.%d' % ver


_DEFAULT_API_VERSION = _format_version(DEFAULT_API_VERSION)


def error_response(exc, code=500):
    res = flask.jsonify(error={'message': str(exc)})
    res.status_code = code
    LOG.debug('Returning error to client: %s', exc)
    return res


def convert_exceptions(func):
    @six.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except utils.Error as exc:
            return error_response(exc, exc.http_code)
        except werkzeug.exceptions.HTTPException as exc:
            return error_response(exc, exc.code or 400)
        except Exception as exc:
            LOG.exception('Internal server error')
            msg = _('Internal server error')
            if CONF.debug:
                msg += ' (%s): %s' % (exc.__class__.__name__, exc)
            return error_response(msg)

    return wrapper


@app.before_request
def check_api_version():
    requested = _get_version()

    if requested < MINIMUM_API_VERSION or requested > CURRENT_API_VERSION:
        return error_response(_('Unsupported API version %(requested)s, '
                                'supported range is %(min)s to %(max)s') %
                              {'requested': _format_version(requested),
                               'min': _format_version(MINIMUM_API_VERSION),
                               'max': _format_version(CURRENT_API_VERSION)},
                              code=406)


@app.after_request
def add_version_headers(res):
    res.headers[conf_opts.MIN_VERSION_HEADER] = '%s.%s' % MINIMUM_API_VERSION
    res.headers[conf_opts.MAX_VERSION_HEADER] = '%s.%s' % CURRENT_API_VERSION
    return res


def create_link_object(urls):
    links = []
    for url in urls:
        links.append({"rel": "self",
                      "href": os.path.join(flask.request.url_root, url)})
    return links


def generate_resource_data(resources):
    data = []
    for resource in resources:
        item = {}
        item['name'] = str(resource).split('/')[-1]
        item['links'] = create_link_object([str(resource)[1:]])
        data.append(item)
    return data


def generate_introspection_status(node):
    """Return a dict representing current node status.

    :param node: a NodeInfo instance
    :return: dictionary
    """
    started_at = node.started_at.isoformat()
    finished_at = node.finished_at.isoformat() if node.finished_at else None

    status = {}
    status['uuid'] = node.uuid
    status['finished'] = bool(node.finished_at)
    status['state'] = node.state
    status['started_at'] = started_at
    status['finished_at'] = finished_at
    status['error'] = node.error
    status['links'] = create_link_object(
        ["v%s/introspection/%s" % (CURRENT_API_VERSION[0], node.uuid)])
    return status


def api(path, is_public_api=False, rule=None, verb_to_rule_map=None,
        **flask_kwargs):
    """Decorator to wrap api methods.

    Performs flask routing, exception conversion,
    generation of oslo context for request and API access policy enforcement.

    :param path: flask app route path
    :param is_public_api: whether this API path should be treated
                          as public, with minimal access enforcement
    :param rule: API access policy rule to enforce.
                 If rule is None, the 'default' policy rule will be enforced,
                 which is "deny all" if not overridden in policy confif file.
    :param verb_to_rule_map: if both rule and this are given,
                             defines mapping between http verbs (uppercase)
                             and strings to format the 'rule' string with
    :param kwargs: all the rest kwargs are passed to flask app.route
    """
    def outer(func):
        @app.route(path, **flask_kwargs)
        @convert_exceptions
        @six.wraps(func)
        def wrapper(*args, **kwargs):
            flask.request.context = context.RequestContext.from_environ(
                flask.request.environ,
                is_public_api=is_public_api)
            if verb_to_rule_map and rule:
                policy_rule = rule.format(
                    verb_to_rule_map[flask.request.method.upper()])
            else:
                policy_rule = rule
            utils.check_auth(flask.request, rule=policy_rule)
            return func(*args, **kwargs)
        return wrapper
    return outer


@api('/', rule='introspection', is_public_api=True, methods=['GET'])
def api_root():
    versions = [
        {
            "status": "CURRENT",
            "id": '%s.%s' % CURRENT_API_VERSION,
        },
    ]

    for version in versions:
        version['links'] = create_link_object(
            ["v%s" % version['id'].split('.')[0]])

    return flask.jsonify(versions=versions)


@api('/<version>', rule='introspection:version', is_public_api=True,
     methods=['GET'])
def version_root(version):
    pat = re.compile(r'^\/%s\/[^\/]*?$' % version)

    resources = []
    for url in app.url_map.iter_rules():
        if pat.match(str(url)):
            resources.append(url)

    if not resources:
        raise utils.Error(_('Version not found.'), code=404)

    return flask.jsonify(resources=generate_resource_data(resources))


@api('/v1/continue', rule="introspection:continue", is_public_api=True,
     methods=['POST'])
def api_continue():
    data = flask.request.get_json(force=True)
    if not isinstance(data, dict):
        raise utils.Error(_('Invalid data: expected a JSON object, got %s') %
                          data.__class__.__name__)

    logged_data = {k: (v if k not in _LOGGING_EXCLUDED_KEYS else '<hidden>')
                   for k, v in data.items()}
    LOG.debug("Received data from the ramdisk: %s", logged_data,
              data=data)

    return flask.jsonify(process.process(data))


# TODO(sambetts) Add API discovery for this endpoint
@api('/v1/introspection/<node_id>',
     rule="introspection:{}",
     verb_to_rule_map={'GET': 'status', 'POST': 'start'},
     methods=['GET', 'POST'])
def api_introspection(node_id):
    if flask.request.method == 'POST':
        args = flask.request.args

        manage_boot = args.get('manage_boot', 'True')
        try:
            manage_boot = strutils.bool_from_string(manage_boot, strict=True)
        except ValueError:
            raise utils.Error(_('Invalid boolean value for manage_boot: %s') %
                              manage_boot, code=400)

        if manage_boot and not CONF.can_manage_boot:
            raise utils.Error(_('Managed boot is requested, but this '
                                'installation cannot manage boot ('
                                '(can_manage_boot set to False)'),
                              code=400)
        client = rpc.get_client()
        client.call({}, 'do_introspection', node_id=node_id,
                    manage_boot=manage_boot,
                    token=flask.request.headers.get('X-Auth-Token'))
        return '', 202
    else:
        node_info = node_cache.get_node(node_id)
        return flask.json.jsonify(generate_introspection_status(node_info))


@api('/v1/introspection', rule='introspection:status', methods=['GET'])
def api_introspection_statuses():
    nodes = node_cache.get_node_list(
        marker=api_tools.marker_field(),
        limit=api_tools.limit_field(default=CONF.api_max_limit)
    )
    data = {
        'introspection': [generate_introspection_status(node)
                          for node in nodes]
    }
    return flask.json.jsonify(data)


@api('/v1/introspection/<node_id>/abort', rule="introspection:abort",
     methods=['POST'])
def api_introspection_abort(node_id):
    client = rpc.get_client()
    client.call({}, 'do_abort', node_id=node_id,
                token=flask.request.headers.get('X-Auth-Token'))
    return '', 202


@api('/v1/introspection/<node_id>/data', rule="introspection:data",
     methods=['GET'])
def api_introspection_data(node_id):
    try:
        if not uuidutils.is_uuid_like(node_id):
            node = ir_utils.get_node(node_id, fields=['uuid'])
            node_id = node.uuid
        res = process.get_introspection_data(node_id)
        return res, 200, {'Content-Type': 'application/json'}
    except utils.IntrospectionDataStoreDisabled:
        return error_response(_('Inspector is not configured to store data. '
                                'Set the [processing]store_data '
                                'configuration option to change this.'),
                              code=404)


@api('/v1/introspection/<node_id>/data/unprocessed',
     rule="introspection:reapply", methods=['POST'])
def api_introspection_reapply(node_id):
    if flask.request.content_length:
        return error_response(_('User data processing is not '
                                'supported yet'), code=400)
    if not uuidutils.is_uuid_like(node_id):
        node = ir_utils.get_node(node_id, fields=['uuid'])
        node_id = node.uuid
    client = rpc.get_client()
    client.call({}, 'do_reapply', node_uuid=node_id)
    return '', 202


def rule_repr(rule, short):
    result = rule.as_dict(short=short)
    result['links'] = [{
        'href': flask.url_for('api_rule', uuid=result['uuid']),
        'rel': 'self'
    }]
    return result


@api('/v1/rules',
     rule="introspection:rule:{}",
     verb_to_rule_map={'GET': 'get', 'POST': 'create', 'DELETE': 'delete'},
     methods=['GET', 'POST', 'DELETE'])
def api_rules():
    if flask.request.method == 'GET':
        res = [rule_repr(rule, short=True) for rule in rules.get_all()]
        return flask.jsonify(rules=res)
    elif flask.request.method == 'DELETE':
        rules.delete_all()
        return '', 204
    else:
        body = flask.request.get_json(force=True)
        if body.get('uuid') and not uuidutils.is_uuid_like(body['uuid']):
            raise utils.Error(_('Invalid UUID value'), code=400)

        rule = rules.create(conditions_json=body.get('conditions', []),
                            actions_json=body.get('actions', []),
                            uuid=body.get('uuid'),
                            description=body.get('description'))

        response_code = (200 if _get_version() < (1, 6) else 201)
        return flask.make_response(
            flask.jsonify(rule_repr(rule, short=False)), response_code)


@api('/v1/rules/<uuid>',
     rule="introspection:rule:{}",
     verb_to_rule_map={'GET': 'get', 'DELETE': 'delete'},
     methods=['GET', 'DELETE'])
def api_rule(uuid):
    if flask.request.method == 'GET':
        rule = rules.get(uuid)
        return flask.jsonify(rule_repr(rule, short=False))
    else:
        rules.delete(uuid)
        return '', 204


@app.errorhandler(404)
def handle_404(error):
    return error_response(error, code=404)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handling introspection data from the ramdisk."""

import copy
import datetime
import os

from oslo_config import cfg
from oslo_serialization import base64
from oslo_utils import excutils
from oslo_utils import timeutils

from ironic_inspector.common.i18n import _
from ironic_inspector.common import ironic as ir_utils
from ironic_inspector import introspection_state as istate
from ironic_inspector import node_cache
from ironic_inspector.plugins import base as plugins_base
from ironic_inspector.pxe_filter import base as pxe_filter
from ironic_inspector import rules
from ironic_inspector import utils

CONF = cfg.CONF

LOG = utils.getProcessingLogger(__name__)

_STORAGE_EXCLUDED_KEYS = {'logs'}


def _store_logs(introspection_data, node_info):
    logs = introspection_data.get('logs')
    if not logs:
        LOG.warning('No logs were passed by the ramdisk',
                    data=introspection_data, node_info=node_info)
        return

    if not CONF.processing.ramdisk_logs_dir:
        LOG.warning('Failed to store logs received from the ramdisk '
                    'because ramdisk_logs_dir configuration option '
                    'is not set',
                    data=introspection_data, node_info=node_info)
        return

    fmt_args = {
        'uuid': node_info.uuid if node_info is not None else 'unknown',
        'mac': (utils.get_pxe_mac(introspection_data) or
                'unknown').replace(':', ''),
        'dt': datetime.datetime.utcnow(),
        'bmc': (utils.get_ipmi_address_from_data(introspection_data) or
                'unknown')
    }

    file_name = CONF.processing.ramdisk_logs_filename_format.format(**fmt_args)

    try:
        if not os.path.exists(CONF.processing.ramdisk_logs_dir):
            os.makedirs(CONF.processing.ramdisk_logs_dir)
        with open(os.path.join(CONF.processing.ramdisk_logs_dir, file_name),
                  'wb') as fp:
            fp.write(base64.decode_as_bytes(logs))
    except EnvironmentError:
        LOG.exception('Could not store the ramdisk logs',
                      data=introspection_data, node_info=node_info)
    else:
        LOG.info('Ramdisk logs were stored in file %s', file_name,
                 data=introspection_data, node_info=node_info)


def _find_node_info(introspection_data, failures):
    try:
        return node_cache.find_node(
            bmc_address=utils.get_ipmi_address_from_data(introspection_data),
            mac=utils.get_valid_macs(introspection_data))
    except utils.NotFoundInCacheError as exc:
        not_found_hook = plugins_base.node_not_found_hook_manager()
        if not_found_hook is None:
            failures.append(_('Look up error: %s') % exc)
            return

        LOG.debug('Running node_not_found_hook %s',
                  CONF.processing.node_not_found_hook,
                  data=introspection_data)

        # NOTE(sambetts): If not_found_hook is not none it means that we were
        # unable to find the node in the node cache and there is a node not
        # found hook defined so we should try to send the introspection data
        # to that hook to generate the node info before bubbling up the error.
        try:
            node_info = not_found_hook.driver(introspection_data)
            if node_info:
                return node_info
            failures.append(_("Node not found hook returned nothing"))
        except Exception as exc:
            failures.append(_("Node not found hook failed: %s") % exc)
    except utils.Error as exc:
        failures.append(_('Look up error: %s') % exc)


def _run_pre_hooks(introspection_data, failures):
    hooks = plugins_base.processing_hooks_manager()
    for hook_ext in hooks:
        LOG.debug('Running pre-processing hook %s', hook_ext.name,
                  data=introspection_data)
        # NOTE(dtantsur): catch exceptions, so that we have changes to update
        # node introspection status after look up
        try:
            hook_ext.obj.before_processing(introspection_data)
        except utils.Error as exc:
            LOG.error('Hook %(hook)s failed, delaying error report '
                      'until node look up: %(error)s',
                      {'hook': hook_ext.name, 'error': exc},
                      data=introspection_data)
            failures.append('Preprocessing hook %(hook)s: %(error)s' %
                            {'hook': hook_ext.name, 'error': exc})
        except Exception as exc:
            LOG.exception('Hook %(hook)s failed, delaying error report '
                          'until node look up: %(error)s',
                          {'hook': hook_ext.name, 'error': exc},
                          data=introspection_data)
            failures.append(_('Unexpected exception %(exc_class)s during '
                              'preprocessing in hook %(hook)s: %(error)s') %
                            {'hook': hook_ext.name,
                             'exc_class': exc.__class__.__name__,
                             'error': exc})


def _filter_data_excluded_keys(data):
    return {k: v for k, v in data.items()
            if k not in _STORAGE_EXCLUDED_KEYS}


def _store_data(node_uuid, data, processed=True):
    introspection_data_manager = plugins_base.introspection_data_manager()
    store = CONF.processing.store_data
    ext = introspection_data_manager[store].obj
    ext.save(node_uuid, data, processed)


def _store_unprocessed_data(node_uuid, data):
    # runs in background
    try:
        _store_data(node_uuid, data, processed=False)
    except Exception:
        LOG.exception('Encountered exception saving unprocessed '
                      'introspection data for node %s', node_uuid, data=data)


def get_introspection_data(uuid, processed=True, get_json=False):
    introspection_data_manager = plugins_base.introspection_data_manager()
    store = CONF.processing.store_data
    ext = introspection_data_manager[store].obj
    return ext.get(uuid, processed=processed, get_json=get_json)


def process(introspection_data):
    """Process data from the ramdisk.

    This function heavily relies on the hooks to do the actual data processing.
    """
    unprocessed_data = copy.deepcopy(introspection_data)
    failures = []
    _run_pre_hooks(introspection_data, failures)
    node_info = _find_node_info(introspection_data, failures)
    if node_info:
        # Locking is already done in find_node() but may be not done in a
        # node_not_found hook
        node_info.acquire_lock()

    if failures or node_info is None:
        msg = _('The following failures happened during running '
                'pre-processing hooks:\n%s') % '\n'.join(failures)
        if node_info is not None:
            node_info.finished(istate.Events.error, error='\n'.join(failures))
        _store_logs(introspection_data, node_info)
        raise utils.Error(msg, node_info=node_info, data=introspection_data)

    LOG.info('Matching node is %s', node_info.uuid,
             node_info=node_info, data=introspection_data)

    if node_info.finished_at is not None:
        # race condition or introspection canceled
        raise utils.Error(_('Node processing already finished with '
                            'error: %s') % node_info.error,
                          node_info=node_info, code=400)

    # Note(mkovacik): store data now when we're sure that a background
    # thread won't race with other process() or introspect.abort()
    # call
    utils.executor().submit(_store_unprocessed_data, node_info.uuid,
                            unprocessed_data)

    try:
        node = node_info.node()
    except ir_utils.NotFound as exc:
        with excutils.save_and_reraise_exception():
            node_info.finished(istate.Events.error, error=str(exc))
            _store_logs(introspection_data, node_info)

    try:
        result = _process_node(node_info, node, introspection_data)
    except utils.Error as exc:
        node_info.finished(istate.Events.error, error=str(exc))
        with excutils.save_and_reraise_exception():
            _store_logs(introspection_data, node_info)
    except Exception as exc:
        LOG.exception('Unexpected exception during processing')
        msg = _('Unexpected exception %(exc_class)s during processing: '
                '%(error)s') % {'exc_class': exc.__class__.__name__,
                                'error': exc}
        node_info.finished(istate.Events.error, error=msg)
        _store_logs(introspection_data, node_info)
        raise utils.Error(msg, node_info=node_info, data=introspection_data,
                          code=500)

    if CONF.processing.always_store_ramdisk_logs:
        _store_logs(introspection_data, node_info)
    return result


def _run_post_hooks(node_info, introspection_data):
    hooks = plugins_base.processing_hooks_manager()

    for hook_ext in hooks:
        LOG.debug('Running post-processing hook %s', hook_ext.name,
                  node_info=node_info, data=introspection_data)
        hook_ext.obj.before_update(introspection_data, node_info)


@node_cache.fsm_transition(istate.Events.process, reentrant=False)
def _process_node(node_info, node, introspection_data):
    # NOTE(dtantsur): repeat the check in case something changed
    ir_utils.check_provision_state(node)
    _run_post_hooks(node_info, introspection_data)
    _store_data(node_info.uuid, introspection_data)

    ironic = ir_utils.get_client()
    pxe_filter.driver().sync(ironic)

    node_info.invalidate_cache()
    rules.apply(node_info, introspection_data)

    resp = {'uuid': node.uuid}

    utils.executor().submit(_finish, node_info, ironic, introspection_data,
                            power_off=CONF.processing.power_off)

    return resp


@node_cache.triggers_fsm_error_transition()
def _finish(node_info, ironic, introspection_data, power_off=True):
    if power_off:
        LOG.debug('Forcing power off of node %s', node_info.uuid)
        try:
            ironic.node.set_power_state(node_info.uuid, 'off')
        except Exception as exc:
            if node_info.node().provision_state == 'enroll':
                LOG.info("Failed to power off the node in"
                         "'enroll' state, ignoring; error was "
                         "%s", exc, node_info=node_info,
                         data=introspection_data)
            else:
                msg = (_('Failed to power off node %(node)s, check '
                         'its power management configuration: '
                         '%(exc)s') % {'node': node_info.uuid, 'exc':
                                       exc})
                raise utils.Error(msg, node_info=node_info,
                                  data=introspection_data)
        LOG.info('Node powered-off', node_info=node_info,
                 data=introspection_data)

    node_info.finished(istate.Events.finish)
    LOG.info('Introspection finished successfully',
             node_info=node_info, data=introspection_data)


def reapply(node_uuid, data=None):
    """Re-apply introspection steps.

    Re-apply preprocessing, postprocessing and introspection rules on
    stored data.

    :param node_uuid: node UUID
    :raises: utils.Error

    """

    LOG.debug('Processing re-apply introspection request for node '
              'UUID: %s', node_uuid)
    node_info = node_cache.get_node(node_uuid, locked=False)
    if not node_info.acquire_lock(blocking=False):
        # Note (mkovacik): it should be sufficient to check data
        # presence & locking. If either introspection didn't start
        # yet, was in waiting state or didn't finish yet, either data
        # won't be available or locking would fail
        raise utils.Error(_('Node locked, please, try again later'),
                          node_info=node_info, code=409)

    utils.executor().submit(_reapply, node_info, data)


def _reapply(node_info, introspection_data=None):
    # runs in background
    node_info.started_at = timeutils.utcnow()
    node_info.commit()

    try:
        ironic = ir_utils.get_client()
    except Exception as exc:
        msg = _('Encountered an exception while getting the Ironic client: '
                '%s') % exc
        LOG.error(msg, node_info=node_info, data=introspection_data)
        node_info.finished(istate.Events.error, error=msg)
        return

    try:
        _reapply_with_data(node_info, introspection_data)
    except Exception as exc:
        msg = (_('Failed reapply for node %(node)s, Error: '
                 '%(exc)s') % {'node': node_info.uuid, 'exc': exc})
        LOG.error(msg, node_info=node_info, data=introspection_data)
        return

    _finish(node_info, ironic, introspection_data,
            power_off=False)

    LOG.info('Successfully reapplied introspection on stored '
             'data', node_info=node_info, data=introspection_data)


@node_cache.fsm_event_before(istate.Events.reapply)
@node_cache.triggers_fsm_error_transition()
def _reapply_with_data(node_info, introspection_data):
    failures = []
    _run_pre_hooks(introspection_data, failures)
    if failures:
        raise utils.Error(_('Pre-processing failures detected reapplying '
                            'introspection on stored data:\n%s') %
                          '\n'.join(failures), node_info=node_info)

    _run_post_hooks(node_info, introspection_data)
    _store_data(node_info.uuid, introspection_data)
    node_info.invalidate_cache()
    rules.apply(node_info, introspection_data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_service import service
from oslo_service import wsgi

from ironic_inspector import main as app
from ironic_inspector import utils

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class WSGIService(service.Service):
    """Provides ability to launch API from wsgi app."""

    def __init__(self):
        self.app = app.app
        self.server = wsgi.Server(CONF, 'ironic_inspector',
                                  self.app,
                                  host=CONF.listen_address,
                                  port=CONF.listen_port,
                                  use_ssl=CONF.use_ssl)

    def _init_middleware(self):
        """Initialize WSGI middleware.

        :returns: None
        """
        if CONF.auth_strategy != 'noauth':
            utils.add_auth_middleware(self.app)
        else:
            LOG.warning('Starting unauthenticated, please check'
                        ' configuration')
        utils.add_cors_middleware(self.app)

    def start(self):
        """Start serving this service using loaded configuration.

        :returns: None
        """
        self._init_middleware()
        self.server.start()

    def stop(self):
        """Stop serving this API.

        :returns: None
        """
        self.server.stop()

    def wait(self):
        """Wait for the service to stop serving this API.

        :returns: None
        """
        self.server.wait()

    def reset(self):
        """Reset server greenpool size to default.

        :returns: None
        """
        self.server.reset()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Migrate introspected data between Swift and database."""

from __future__ import print_function

import sys

from oslo_config import cfg
from oslo_log import log
from oslo_utils import encodeutils
import six

from ironic_inspector.common.i18n import _
from ironic_inspector.conf import opts
from ironic_inspector import node_cache
from ironic_inspector.plugins import base as plugins_base
from ironic_inspector import utils

LOG = log.getLogger(__name__)
CONF = cfg.CONF

_AVAILABLE_STORAGES = [('database', _('The database storage backend')),
                       ('swift', _('The Swift storage backend'))]
_OPTS = [
    cfg.StrOpt('from',
               dest='source_storage',
               required=True,
               choices=_AVAILABLE_STORAGES,
               help=_('The source storage where the introspected data will be '
                      'read from.')),
    cfg.StrOpt('to',
               dest='target_storage',
               required=True,
               choices=_AVAILABLE_STORAGES,
               help=_('The target storage where the introspected data will be '
                      'saved to.'))
]

# Migration result
RESULT_NOCONTENT = 'no content'
RESULT_FAILED = 'failed'
RESULT_SUCCESS = 'success'


def _setup_logger(args=None):
    args = [] if args is None else args
    log.register_options(CONF)
    opts.set_config_defaults()
    opts.parse_args(args)
    log.setup(CONF, 'ironic_inspector')


class MigrationTool(object):

    def _migrate_one(self, node, processed):
        LOG.debug('Starting to migrate introspection data for node '
                  '%(node)s (processed %(processed)s)',
                  {'node': node.uuid, 'processed': processed})
        try:
            data = self.ext_src.get(node.uuid, processed=processed,
                                    get_json=True)
            if not data:
                return RESULT_NOCONTENT
            self.ext_tgt.save(node.uuid, data, processed=processed)
        except Exception as e:
            LOG.error('Migrate introspection data failed for node '
                      '%(node)s (processed %(processed)s), error: '
                      '%(error)s', {'node': node.uuid, 'processed': processed,
                                    'error': e})
            return RESULT_FAILED

        return RESULT_SUCCESS

    def main(self):
        CONF.register_cli_opts(_OPTS)
        _setup_logger(sys.argv[1:])

        if CONF.source_storage == CONF.target_storage:
            raise utils.Error(_('Source and destination can not be the same.'))

        introspection_data_manager = plugins_base.introspection_data_manager()
        self.ext_src = introspection_data_manager[CONF.source_storage].obj
        self.ext_tgt = introspection_data_manager[CONF.target_storage].obj

        nodes = node_cache.get_node_list()
        migration_list = [(n, p) for n in nodes for p in [True, False]]
        failed_records = []
        for node, processed in migration_list:
            result = self._migrate_one(node, processed)
            if result == RESULT_FAILED:
                failed_records.append((node.uuid, processed))

        msg = ('Finished introspection data migration, total records: %d. '
               % len(migration_list))
        if failed_records:
            msg += 'Failed to migrate:\n' + '\n'.join([
                '%s(processed=%s)' % (record[0], record[1])
                for record in failed_records])
        elif len(migration_list) > 0:
            msg += 'all records are migrated successfully.'
        print(msg)


def main():

    try:
        MigrationTool().main()
    except KeyboardInterrupt:
        print(_("... terminating migration tool"), file=sys.stderr)
        return 130
    except Exception as e:
        print(encodeutils.safe_encode(six.text_type(e)), file=sys.stderr)
        return 1


if __name__ == '__main__':
    sys.exit(main())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from ironic_inspector.common.i18n import _
from ironic_inspector.common import keystone


SWIFT_GROUP = 'swift'
SERVICE_TYPE = 'object-store'


_OPTS = [
    cfg.IntOpt('max_retries',
               default=2,
               help=_('Maximum number of times to retry a Swift request, '
                      'before failing.')),
    cfg.IntOpt('delete_after',
               default=0,
               help=_('Number of seconds that the Swift object will last '
                      'before being deleted. (set to 0 to never delete the '
                      'object).')),
    cfg.StrOpt('container',
               default='ironic-inspector',
               help=_('Default Swift container to use when creating '
                      'objects.')),
    cfg.StrOpt('os_service_type',
               default='object-store',
               help=_('Swift service type.'),
               deprecated_for_removal=True,
               deprecated_reason=_('Use [swift]/service_type option '
                                   'to set specific service type')),
    cfg.StrOpt('os_endpoint_type',
               default='internalURL',
               help=_('Swift endpoint type.'),
               deprecated_for_removal=True,
               deprecated_reason=_('Use [swift]/valid_interfaces option '
                                   'to specify endpoint interfaces.')),
    cfg.StrOpt('os_region',
               help=_('Keystone region to get endpoint for.'),
               deprecated_for_removal=True,
               deprecated_reason=_("Use [swift]/region_name option to "
                                   "configure region."))
]


def register_opts(conf):
    conf.register_opts(_OPTS, SWIFT_GROUP)
    keystone.register_auth_opts(SWIFT_GROUP, SERVICE_TYPE)


def list_opts():
    return keystone.add_auth_options(_OPTS, SERVICE_TYPE)
